import java.io.*;
import java.util.*;

public class speechNet extends CANTNet {
  private LineNumberReader inputFile;
  public int stepsPerInput = 3;

  public speechNet(){
  }

  public speechNet(String name,int cols, int rows,int topology){
    super(name,cols,rows,topology);
    if (name.compareTo("BaseNet") == 0) {
      openInputFile();
    }
    else setCurrentPattern(0);
  }

  public void readBetweenAllNets() {
    System.out.println("speechNet read Between Connections Called");
    int netsChecked = 0;
    Enumeration eNum = CANT23.nets.elements();
    speechNet net = (speechNet) eNum.nextElement();

    speechNet baseNet = net;
    speechNet phonemeNet = net;

    do {
      System.out.println(net.getName());
      if (net.getName().compareTo("BaseNet") == 0)
        baseNet = net;
      else if (net.getName().compareTo("PhonemeNet") == 0)
        phonemeNet = net;
      else
        System.out.println(net.getName() + " missed net in connect all");
      netsChecked++;
      if (netsChecked < 2)
        net = (speechNet) eNum.nextElement();
    } while (netsChecked < 2);
    baseNet.readConnectTo(phonemeNet);
  }

  private String getWordFileName(int epoch, boolean test) {
    String  fileName = "norm.txt";
    //String  fileName = "form.txt";
    int offset;
    if (test) {
      offset=speechCANT23.testSuite*2;
    }
    else { //train
      offset=speechCANT23.trainSuite*2;
    }

    if ((epoch % 6 ) >= 3) offset = offset - 1;
      fileName = offset+fileName;
    if ((epoch % 3 ) == 0)
      fileName = "loot"+fileName;
    else if ((epoch % 3 ) == 1)
      fileName = "father"+fileName;
    else if ((epoch % 3 ) == 2)
      fileName = "keep"+fileName;
    return fileName;      
  }

  public int time = 0;
  public void openInputFile() {
    DataInputStream dIS;
    InputStreamReader inputSR;
    String wordFile = "";

    if (speechCANT23.epoch == (120*6) + 1) 
	{
        speechCANT23.trainSuite++;
        if (speechCANT23.trainSuite > 5) {
            speechCANT23.trainSuite = 1;
            CANT23.seed++;
	  }
	}

    if (speechCANT23.epoch > 120) {//test
      wordFile = getWordFileName(speechCANT23.epoch,true);
      if ((speechCANT23.epoch % 120) == 0) speechCANT23.testSuite++;
    }
    else //train
      wordFile = getWordFileName(speechCANT23.epoch,false);

System.out.println(speechCANT23.epoch  + " " + wordFile);

    inputFile = null;

    try {
      dIS = new DataInputStream(new FileInputStream(wordFile));
      inputSR = new InputStreamReader(dIS);
      inputFile = new LineNumberReader(inputSR);
      String inputLine = inputFile.readLine(); 

      StringTokenizer tokenizedLine = new StringTokenizer(inputLine);

      String nextTokenString = tokenizedLine.nextToken();
      int freqs = Integer.parseInt(nextTokenString);
      nextTokenString = tokenizedLine.nextToken();
      time = Integer.parseInt(nextTokenString);
    }

    catch (IOException e) {
      System.err.println("word file not opened properly\n" + e.toString());
      System.exit(1);
    }
  }

  public void closeInputFile() {
  }

  //destructor
  protected void finalize() throws Throwable
  {
    //do finalization here
    super.finalize(); //not necessary if extending Object.
    if (getName().compareTo("BaseNet") == 0) inputFile.close();
  } 

  public void connectBaseToPhoneme(speechNet phonemeNet) {
    for (int fromNeuron = 0; fromNeuron < getSize(); fromNeuron++) {
      for (int synapse = 0; synapse < 20; synapse++) {
        int toNeuron = CANT23.random.nextInt(phonemeNet.getSize());
        neurons[fromNeuron].addConnection(phonemeNet.neurons[toNeuron], 0.02);
      }
    }
  }


  //-----------------------------Set Up Patterns
  private int convertFreqValToNumNeurons(Double freqVal) {
    int retVal = (int) Math.round(freqVal/10000);
    //System.out.println(freqVal + " " + retVal);
    return retVal;
  }

  private boolean patternElementIsNew(int cPoints, int newEl, 
    int[] patternPoints) {
    for (int i = 0; i < cPoints; i++) {
      if (patternPoints[i] == newEl) 
	{
        //System.out.println(newEl + " redundant");
        return false;
	}
    }
    return true;
  }

  //Fill the pattern and return the number of items.  return -1 if you 
  //get to the end of the file
  private int readSpeechLine(int[] patternPoints) {
    StringTokenizer tokenizedLine;
    int cPoints = 0;

    try {
      String inputLine = inputFile.readLine();
      //System.out.println(inputLine);
      if (inputLine == null) return -1;
      tokenizedLine = new StringTokenizer(inputLine);

//System.out.println(" newline ");
      //read each of the frequency values
      for (int inputs = 0; inputs < 20 ; inputs++) {
       if(!tokenizedLine.hasMoreTokens()) return -1;
        String nextTokenString = tokenizedLine.nextToken();
        int numNeurons = Integer.parseInt(nextTokenString);
//if ((inputs == 5) || (inputs == 13) || (inputs == 35))
//System.out.print(" " + numNeurons);

        for (int neuronsToAdd = 0; neuronsToAdd < numNeurons; neuronsToAdd++)
	{
          int neuronToAdd =  CANT23.random.nextInt(20);
          neuronToAdd += (inputs*20);
          if (patternElementIsNew(cPoints,neuronToAdd,patternPoints)) {
  	    patternPoints[cPoints] = neuronToAdd;
            cPoints++;
          }
        }
      }   
      setNeuronsToStimulate(cPoints);
    } 
    catch (IOException e) {
      System.err.println("word readline problem\n" + e.toString());
      System.exit(1);
    }
    return cPoints;
  }

  //From the frequency file, get the next pattern that roughly reflects
  //the current neural firing pattern.
  private CANTPattern readNewPattern(String fileName) {
    CANTPattern readPattern;
    int[] patternPoints = new int[2560];
    int cPoints = 0;
    int number;

    cPoints = readSpeechLine(patternPoints);
    // System.out.println(" read New Pattern " + getTotalPatterns());
    //If we get -1, there is no more speech to read, so stop stimulating.
    if (cPoints == -1) { 
      readPattern = new CANTPattern(this, "scene", 1, 0, patternPoints);
      setNeuronsToStimulate(0);
      //CANT23.isRunning = false;
      }
    else 
      readPattern = new CANTPattern(this, "scene", 1, cPoints, patternPoints);
    // readPattern.print();
    return readPattern;
  }

  private void modifyPattern(int patternNumber, String fileName) {
    // Patterns aren't actually patterns but a vector of CANTPatterns.
    // We're going to get the CANTPattern at patternNumber.
    CANTPattern newPattern = readNewPattern(fileName);
    // System.out.println(" read New Picture Pattern " + fileName);
    // newPattern.print();

    // and re setElementAt
    patterns.setElementAt(newPattern, patternNumber);
  }

  //need this to subclass experiment
  public void changePattern(int cantStep)
  {
    if (getName().compareTo("BaseNet") == 0) {
      if ((cantStep %stepsPerInput) == 0) {//Each input ~85ms of input, 
                                           //so go for 9 cycles
        modifyPattern(1,"stub");
        setCurrentPattern(1);
        //((CANTPattern)patterns.get(getCurrentPattern())).print();
      }
    }
    else
      {
	  //      System.out.println(" rearrange " + getName());
      ((CANTPattern)patterns.get(getCurrentPattern())).arrange(50);
      }
  }

  //Only set external activation on the base net
    /*  public void setExternalActivation(int cantStep)
  {
    if (getName().compareTo("BaseNet") == 0)
      super.setExternalActivation(cantStep);
  }
    */
  
  public void runAllOneStep(int CANTStep) {
    //This series of loops is really chaotic, but I needed to
    //get all of the propogation done in each net in step.
    speechCANT23.runOneStepStart();
	
    Enumeration eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      //net.runOneStep(CANTStep);
      net.changePattern(CANTStep);
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.setExternalActivation(CANTStep);
    }
      //net.propogateChange();  
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.spontaneousActivate();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.setNeuronsFired();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.setDecay ();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.spreadActivation();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.setFatigue();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.learn();
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      net.cantFrame.runOneStep(CANTStep+1);
    }
    eNum = CANT23.nets.elements();
    while (eNum.hasMoreElements()) {
      speechNet net = (speechNet)eNum.nextElement();
      if (net.recordingActivation) net.setMeasure(CANTStep); 	  
//if (net.getName().compareTo("VerbNet") == 0)   System.out.println(net.neurons[0].getFatigue() +   " verb Neuron " + net.neurons[0].getActivation());
    }
  }


  public CANTNet getNewNet(String name,int cols, int rows,int topology){
    speechNet net = new speechNet (name,cols,rows,topology);
    return (net);
  } 

  private void setPhonemeConnections(int startNeuron, int endNeuron) {
      System.out.println("speech set conn ");
    int currentFromNeuron;
    int N1 = -1,N2 = -1,N3 = -1,N4 = -1;
    int cNeurons = size();


    //For Each Neuron
    for (currentFromNeuron = startNeuron ; currentFromNeuron < endNeuron;
         currentFromNeuron ++) {

      if (neurons[currentFromNeuron].isInhibitory()){
        setConnectionsRandomly(currentFromNeuron,40,-0.1);
      }
      else      {
        //Get it's neighbors Make N1 a long distance connection
        N1 = (int)(N1 + (CANT23.random.nextFloat() * size()))%size();
        N2 = getRightNeighbor(currentFromNeuron);
        N3 = getTopNeighbor(currentFromNeuron);
        N4 = getBottomNeighbor(currentFromNeuron);
        recursiveSetConnections(currentFromNeuron,N1,1);
        recursiveSetConnections(currentFromNeuron,N2,1);
        recursiveSetConnections(currentFromNeuron,N3,1);
        recursiveSetConnections(currentFromNeuron,N4,1);
      }
    }
  }

  private void setPhonemeTopology() {
    setPhonemeConnections(0,size());
    //setConnections(0,size());  Everything distance biased
    for (int neuronNum=0; neuronNum<size(); neuronNum++)  
      {
        int fromCA = neuronNum/200;
        for (int synapseNum=0;synapseNum < 
          neurons[neuronNum].getCurrentSynapses(); synapseNum++) 
          {
	    int toNeuron = neurons[neuronNum].synapses[synapseNum].toNeuron.id;
            int toCA = toNeuron/200;
	    if (neurons[neuronNum].getInhibitory()) 
              {
	      if (fromCA == toCA) 
	        neurons[neuronNum].synapses[synapseNum].setWeight(-0.1);
	      else 
	        neurons[neuronNum].synapses[synapseNum].setWeight(-1.0);
	      }
            else
	      {
	      if (fromCA == toCA) 
	        neurons[neuronNum].synapses[synapseNum].setWeight(0.2);
	      else 
	        neurons[neuronNum].synapses[synapseNum].setWeight(0.01);
	      }

	  }
      }
  }

  public void initializeNeurons() {
    //set up topologies.
    createNeurons();
															  
    if (topology == 1){
      System.out.println("input topology ");
    }
    else if (topology == 2){
      System.out.println("phoneme topology ");
      setPhonemeTopology();
    }
    else 
      System.out.println("bad toppology specified "+ topology);
  }

  public void kludge () {
    System.out.println("kludge " + speechCANT23.experiment.aCount + " " +
      speechCANT23.experiment.eCount + " " +
      speechCANT23.experiment.uCount);
  }
  
  public void measure(int currentStep) {
    System.out.println("measure " + neurons[0].getActivation() + " " + 
      neurons[0].getFired() + " " + currentStep);
  }

  public void printWeights() {
    int aInhibSynapses = 0;
    int eInhibSynapses = 0 ; 
    int uInhibSynapses = 0;
    int aExciteSynapses = 0;
    int eExciteSynapses = 0;
    int uExciteSynapses = 0;
    double aInhibWeight = 0.0;
    double eInhibWeight = 0.0;
    double uInhibWeight = 0.0;
    double aExciteWeight = 0.0;
    double eExciteWeight = 0.0;
    double uExciteWeight = 0.0;
    for (int neuron = 0; neuron< getSize(); neuron++) {
      for (int synapse = 0; synapse< neurons[neuron].getCurrentSynapses(); 
        synapse++) {
        if (neurons[neuron].synapses[synapse].toNeuron.parentNet.getName().
	    compareTo("BaseNet") == 0); //ignore synapses to basenet
        else if (neurons[neuron].synapses[synapse].toNeuron.id < 200){
          if (neurons[neuron].isInhibitory()) {
	    aInhibSynapses ++;
            aInhibWeight += neurons[neuron].synapses[synapse].weight;
	  }
          else {
	    aExciteSynapses ++;
            aExciteWeight += neurons[neuron].synapses[synapse].weight;
	  }
        }

        else if (neurons[neuron].synapses[synapse].toNeuron.id < 400){
          if (neurons[neuron].isInhibitory()) {
	    eInhibSynapses ++;
            eInhibWeight += neurons[neuron].synapses[synapse].weight;
	  }
          else {
	    eExciteSynapses ++;
            eExciteWeight += neurons[neuron].synapses[synapse].weight;
	  }
	}

        else {
          if (neurons[neuron].isInhibitory()) {
	    uInhibSynapses ++;
            uInhibWeight += neurons[neuron].synapses[synapse].weight;
	  }
          else {
	    uExciteSynapses ++;
            uExciteWeight += neurons[neuron].synapses[synapse].weight;
	  }
        }
      }
    }
    System.out.println("a " +  aExciteWeight + " " + aExciteSynapses + " " +
                        aExciteWeight/aExciteSynapses);
    System.out.println("e " +  eExciteWeight + " " + eExciteSynapses + " " +
                        eExciteWeight/eExciteSynapses);
    System.out.println("u " +  uExciteWeight + " " + uExciteSynapses + " " +
                        uExciteWeight/uExciteSynapses);

  }


} //end class